@@ -9,6 +9,8 @@ import (
 	"strings"
 	"time"
 
+	"k8s.io/klog"
+
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/api/resource"
@@ -20,6 +22,7 @@ import (
 	"github.com/onsi/ginkgo/extensions/table"
 	. "github.com/onsi/gomega"
 
+	performancev2 "github.com/openshift-kni/performance-addon-operators/api/v2"
 	testutils "github.com/openshift-kni/performance-addon-operators/functests/utils"
 	testclient "github.com/openshift-kni/performance-addon-operators/functests/utils/client"
 	"github.com/openshift-kni/performance-addon-operators/functests/utils/discovery"
@@ -27,13 +30,13 @@ import (
 	"github.com/openshift-kni/performance-addon-operators/functests/utils/nodes"
 	"github.com/openshift-kni/performance-addon-operators/functests/utils/pods"
 	"github.com/openshift-kni/performance-addon-operators/functests/utils/profiles"
-	performancev1 "github.com/openshift-kni/performance-addon-operators/pkg/apis/performance/v1"
 	"github.com/openshift-kni/performance-addon-operators/pkg/controller/performanceprofile/components"
 )
 
+var workerRTNode *corev1.Node
+var profile *performancev2.PerformanceProfile
+
 var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
-	var workerRTNode *corev1.Node
-	var profile *performancev1.PerformanceProfile
 	var balanceIsolated bool
 	var reservedCPU, isolatedCPU string
 	var listReservedCPU, listIsolatedCPU []int
@@ -51,11 +54,9 @@ var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
 		Expect(workerRTNodes).ToNot(BeEmpty())
 		workerRTNode = &workerRTNodes[0]
 		profile, err = profiles.GetByNodeLabels(testutils.NodeSelectorLabels)
-		By(fmt.Sprintf("Checking the profile %s with cpus %#v", profile.Name, profile.Spec.CPU))
-
 		Expect(err).ToNot(HaveOccurred())
-		Expect(profile.Spec.HugePages).ToNot(BeNil())
 
+		By(fmt.Sprintf("Checking the profile %s with cpus %#v", profile.Name, profile.Spec.CPU))
 		balanceIsolated = true
 		if profile.Spec.CPU.BalanceIsolated != nil {
 			balanceIsolated = *profile.Spec.CPU.BalanceIsolated
@@ -82,7 +83,7 @@ var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
 			Expect(capacityCPU - allocatableCPU).To(Equal(int64(len(listReservedCPU))))
 		})
 
-		It("[test_id:27081][crit:high][vendor:cnf-qe@redhat.com][level:acceptance] Verify CPU affinity mask, CPU reservation and CPU isolation on worker node", func() {
+		It("[test_id:37862][crit:high][vendor:cnf-qe@redhat.com][level:acceptance] Verify CPU affinity mask, CPU reservation and CPU isolation on worker node", func() {
 			By("checking isolated CPU")
 			cmd := []string{"cat", "/sys/devices/system/cpu/isolated"}
 			sysIsolatedCpus, err := nodes.ExecCommandOnNode(cmd, workerRTNode)
@@ -147,7 +148,7 @@ var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
 		})
 	})
 
-	Describe("[test_id:27492][crit:high][vendor:cnf-qe@redhat.com][level:acceptance] Verification of cpu manager functionality", func() {
+	Describe("Verification of cpu manager functionality", func() {
 		var testpod *corev1.Pod
 		var discoveryFailed bool
 
@@ -172,11 +173,7 @@ var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
 		})
 
 		AfterEach(func() {
-			err := testclient.Client.Delete(context.TODO(), testpod)
-			Expect(err).ToNot(HaveOccurred())
-
-			err = pods.WaitForDeletion(testpod, 60*time.Second)
-			Expect(err).ToNot(HaveOccurred())
+			deleteTestPod(testpod)
 		})
 
 		table.DescribeTable("Verify CPU usage by stress PODs", func(guaranteed bool) {
@@ -220,81 +217,73 @@ var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
 
 			Expect(cpu).To(BeElementOf(listCPU))
 		},
-			table.Entry("Non-guaranteed POD can work on any CPU", false),
-			table.Entry("Guaranteed POD should work on isolated cpu", true),
+			table.Entry("[test_id:37860] Non-guaranteed POD can work on any CPU", false),
+			table.Entry("[test_id:27492] Guaranteed POD should work on isolated cpu", true),
 		)
 	})
 
 	When("pod runs with the CPU load balancing runtime class", func() {
 		var testpod *corev1.Pod
-		var defaultFlags []int
+		var defaultFlags map[int][]int
 
-		getCPUSchedulingDomainFlags := func(cpu int) ([]int, error) {
-			cmd := []string{"/bin/bash", "-c", fmt.Sprintf("cat /proc/sys/kernel/sched_domain/cpu%d/domain*/flags", cpu)}
+		getCPUsSchedulingDomainFlags := func() (map[int][]int, error) {
+			cmd := []string{"/bin/bash", "-c", "more /proc/sys/kernel/sched_domain/cpu*/domain*/flags | cat"}
 			out, err := nodes.ExecCommandOnNode(cmd, workerRTNode)
 			if err != nil {
 				return nil, err
 			}
 
-			var domainsFlags []int
-			for _, domainFlags := range strings.Split(out, "\n") {
-				flags, err := strconv.Atoi(domainFlags)
+			re, err := regexp.Compile(`/proc/sys/kernel/sched_domain/cpu(\d+)/domain\d+/flags\n:+\n(\d+)`)
+			if err != nil {
+				return nil, err
+			}
+
+			allSubmatch := re.FindAllStringSubmatch(out, -1)
+			cpuToSchedDomains := map[int][]int{}
+			for _, submatch := range allSubmatch {
+				if len(submatch) != 3 {
+					return nil, fmt.Errorf("the sched_domain submatch %v does not have a valid length", submatch)
+				}
+
+				cpu, err := strconv.Atoi(submatch[1])
 				if err != nil {
 					return nil, err
 				}
-				domainsFlags = append(domainsFlags, flags)
+
+				if _, ok := cpuToSchedDomains[cpu]; !ok {
+					cpuToSchedDomains[cpu] = []int{}
+				}
+
+				flags, err := strconv.Atoi(submatch[2])
+				if err != nil {
+					return nil, err
+				}
+
+				cpuToSchedDomains[cpu] = append(cpuToSchedDomains[cpu], flags)
 			}
 
-			sort.Ints(domainsFlags)
-			return domainsFlags, nil
+			// sort sched_domain
+			for cpu := range cpuToSchedDomains {
+				sort.Ints(cpuToSchedDomains[cpu])
+			}
+
+			klog.Infof("Scheduler domains: %v", cpuToSchedDomains)
+			return cpuToSchedDomains, nil
 		}
 
 		BeforeEach(func() {
-			// get the default value for sched_domain flags, we will check the flags only for the CPU 0, because on the
-			// clean system it should be the same for all CPUs
 			var err error
-			defaultFlags, err = getCPUSchedulingDomainFlags(0)
+			defaultFlags, err = getCPUsSchedulingDomainFlags()
 			Expect(err).ToNot(HaveOccurred())
 
-			testpod = pods.GetTestPod()
-			testpod.Annotations = map[string]string{
-				"cpu-load-balancing.crio.io": "true",
+			annotations := map[string]string{
+				"cpu-load-balancing.crio.io": "disable",
 			}
-			testpod.Namespace = testutils.NamespaceTesting
-
-			cpus := resource.MustParse("1")
-			memory := resource.MustParse("256Mi")
-
-			// change pod resource requirements, to change the pod QoS class to guaranteed
-			testpod.Spec.Containers[0].Resources = corev1.ResourceRequirements{
-				Limits: corev1.ResourceList{
-					corev1.ResourceCPU:    cpus,
-					corev1.ResourceMemory: memory,
-				},
-			}
-
-			// use the CPU load balancing runtime class
-			runtimeClassName := components.GetComponentName(profile.Name, components.ComponentNamePrefix)
-			testpod.Spec.RuntimeClassName = &runtimeClassName
-			testpod.Spec.NodeSelector = map[string]string{testutils.LabelHostname: workerRTNode.Name}
+			testpod = getTestPodWithAnnotations(annotations)
 		})
 
 		AfterEach(func() {
-			// it possible that the pod already was deleted as part of the test, in this case we want to skip teardown
-			key := types.NamespacedName{
-				Name:      testpod.Name,
-				Namespace: testpod.Namespace,
-			}
-			err := testclient.Client.Get(context.TODO(), key, testpod)
-			if errors.IsNotFound(err) {
-				return
-			}
-
-			err = testclient.Client.Delete(context.TODO(), testpod)
-			Expect(err).ToNot(HaveOccurred())
-
-			err = pods.WaitForDeletion(testpod, 60*time.Second)
-			Expect(err).ToNot(HaveOccurred())
+			deleteTestPod(testpod)
 		})
 
 		It("[test_id:32646] should disable CPU load balancing for CPU's used by the pod", func() {
@@ -318,36 +307,165 @@ var _ = Describe("[rfe_id:27363][performance] CPU Management", func() {
 			output, err := nodes.ExecCommandOnNode(cmd, workerRTNode)
 			Expect(err).ToNot(HaveOccurred())
 
-			cpu, err := strconv.Atoi(output)
+			cpus, err := cpuset.Parse(output)
 			Expect(err).ToNot(HaveOccurred())
 
 			By("Getting the CPU scheduling flags")
-			flags, err := getCPUSchedulingDomainFlags(cpu)
+			flags, err := getCPUsSchedulingDomainFlags()
 			Expect(err).ToNot(HaveOccurred())
 
 			By("Verifying that the CPU load balancing was disabled")
-			Expect(len(flags)).To(Equal(len(defaultFlags)))
-			// the CPU flags should be almost the same except the LSB that should be disabled
-			for i := range flags {
-				Expect(flags[i]).To(Equal(defaultFlags[i] - 1))
+			for _, cpu := range cpus.ToSlice() {
+				Expect(len(flags[cpu])).To(Equal(len(defaultFlags[cpu])))
+				// the CPU flags should be almost the same except the LSB that should be disabled
+				// see https://github.com/torvalds/linux/blob/0fe5f9ca223573167c4c4156903d751d2c8e160e/include/linux/sched/topology.h#L14
+				// for more information regarding the sched domain flags
+				for i := range flags[cpu] {
+					Expect(flags[cpu][i]).To(Equal(defaultFlags[cpu][i] - 1))
+				}
 			}
 
 			By("Deleting the pod")
-			err = testclient.Client.Delete(context.TODO(), testpod)
-			Expect(err).ToNot(HaveOccurred())
-
-			err = pods.WaitForDeletion(testpod, 60*time.Second)
-			Expect(err).ToNot(HaveOccurred())
+			deleteTestPod(testpod)
 
 			By("Getting the CPU scheduling flags")
-			flags, err = getCPUSchedulingDomainFlags(cpu)
+			flags, err = getCPUsSchedulingDomainFlags()
 			Expect(err).ToNot(HaveOccurred())
 
 			By("Verifying that the CPU load balancing was enabled back")
-			Expect(len(flags)).To(Equal(len(defaultFlags)))
-			// the CPU scheduling flags should be restored to the default values
-			for i := range flags {
-				Expect(flags[i]).To(Equal(defaultFlags[i]))
+			for _, cpu := range cpus.ToSlice() {
+				Expect(len(flags[cpu])).To(Equal(len(defaultFlags[cpu])))
+				// the CPU scheduling flags should be restored to the default values
+				for i := range flags[cpu] {
+					Expect(flags[cpu][i]).To(Equal(defaultFlags[cpu][i]))
+				}
+			}
+		})
+	})
+
+	Describe("Verification that IRQ load balance can be disabled per POD", func() {
+		var testpod *corev1.Pod
+
+		BeforeEach(func() {
+			if profile.Spec.GloballyDisableIrqLoadBalancing != nil && *profile.Spec.GloballyDisableIrqLoadBalancing {
+				Skip("IRQ load balance should be enabled (GloballyDisableIrqLoadBalancing=false), skipping test")
+			}
+		})
+
+		AfterEach(func() {
+			deleteTestPod(testpod)
+		})
+
+		It("[test_id:36364] should disable IRQ balance for CPU where POD is running", func() {
+			By("checking default smp affinity is equal to all active CPUs")
+			defaultSmpAffinitySet, err := nodes.GetDefaultSmpAffinitySet(workerRTNode)
+			Expect(err).ToNot(HaveOccurred())
+
+			onlineCPUsSet, err := nodes.GetOnlineCPUsSet(workerRTNode)
+			Expect(err).ToNot(HaveOccurred())
+
+			Expect(defaultSmpAffinitySet).To(Equal(onlineCPUsSet), fmt.Sprintf("Default SMP Affinity %s should be equal to all active CPUs %s", defaultSmpAffinitySet, onlineCPUsSet))
+
+			By("Running pod with annotations that disable specific CPU from IRQ balancer")
+			annotations := map[string]string{
+				"irq-load-balancing.crio.io": "disable",
+				"cpu-quota.crio.io":          "disable",
+			}
+			testpod = getTestPodWithAnnotations(annotations)
+
+			err = testclient.Client.Create(context.TODO(), testpod)
+			Expect(err).ToNot(HaveOccurred())
+			err = pods.WaitForCondition(testpod, corev1.PodReady, corev1.ConditionTrue, 10*time.Minute)
+			Expect(err).ToNot(HaveOccurred())
+
+			By("Checking that the default smp affinity mask was updated and CPU (where POD is running) isolated")
+			defaultSmpAffinitySet, err = nodes.GetDefaultSmpAffinitySet(workerRTNode)
+			Expect(err).ToNot(HaveOccurred())
+
+			getPsr := []string{"/bin/bash", "-c", "grep Cpus_allowed_list /proc/self/status | awk '{print $2}'"}
+			psr, err := pods.ExecCommandOnPod(testpod, getPsr)
+			Expect(err).ToNot(HaveOccurred())
+			psrSet, err := cpuset.Parse(strings.Trim(string(psr), "\n"))
+			Expect(err).ToNot(HaveOccurred())
+
+			Expect(defaultSmpAffinitySet).To(Equal(onlineCPUsSet.Difference(psrSet)), fmt.Sprintf("Default SMP affinity should not contain isolated CPU %s", psr))
+
+			By("Checking that there are no any active IRQ on isolated CPU")
+			// It may takes some time for the system to reschedule active IRQs
+			Eventually(func() bool {
+				getActiveIrq := []string{"/bin/bash", "-c", "for n in $(find /proc/irq/ -name smp_affinity_list); do echo $(cat $n); done"}
+				activeIrq, err := nodes.ExecCommandOnNode(getActiveIrq, workerRTNode)
+				Expect(err).ToNot(HaveOccurred())
+				for _, irq := range strings.Split(activeIrq, "\n") {
+					irqAffinity, err := cpuset.Parse(irq)
+					Expect(err).ToNot(HaveOccurred())
+					if !irqAffinity.Equals(onlineCPUsSet) && psrSet.IsSubsetOf(irqAffinity) {
+						return false
+					}
+				}
+				return true
+			}, 30*time.Second, 5*time.Second).Should(BeTrue(),
+				fmt.Sprintf("IRQ still active on CPU%s", psr))
+
+			By("Checking that after removing POD default smp affinity is returned back to all active CPUs")
+			deleteTestPod(testpod)
+			defaultSmpAffinitySet, err = nodes.GetDefaultSmpAffinitySet(workerRTNode)
+			Expect(err).ToNot(HaveOccurred())
+
+			Expect(defaultSmpAffinitySet).To(Equal(onlineCPUsSet), fmt.Sprintf("Default SMP Affinity %s should be equal to all active CPUs %s", defaultSmpAffinitySet, onlineCPUsSet))
+		})
+	})
+
+	When("reserved CPUs specified", func() {
+		var testpod *corev1.Pod
+
+		BeforeEach(func() {
+			testpod = pods.GetTestPod()
+			testpod.Namespace = testutils.NamespaceTesting
+			testpod.Spec.NodeSelector = map[string]string{testutils.LabelHostname: workerRTNode.Name}
+
+			err := testclient.Client.Create(context.TODO(), testpod)
+			Expect(err).ToNot(HaveOccurred())
+
+			err = pods.WaitForCondition(testpod, corev1.PodReady, corev1.ConditionTrue, 10*time.Minute)
+			Expect(err).ToNot(HaveOccurred())
+		})
+
+		It("should run infra containers on reserved CPUs", func() {
+			// find used because that crictl does not show infra containers, `runc list` shows them
+			// but you will need somehow to find infra containers ID's
+			podUID := strings.Replace(string(testpod.UID), "-", "_", -1)
+
+			cmd := []string{"/bin/bash", "-c", fmt.Sprintf("find /rootfs/sys/fs/cgroup/cpuset/ -name *%s*", podUID)}
+			podCgroup, err := nodes.ExecCommandOnNode(cmd, workerRTNode)
+			Expect(err).ToNot(HaveOccurred())
+
+			cmd = []string{"/bin/bash", "-c", fmt.Sprintf("find %s -name crio-*", podCgroup)}
+			containersCgroups, err := nodes.ExecCommandOnNode(cmd, workerRTNode)
+			Expect(err).ToNot(HaveOccurred())
+
+			containerID, err := pods.GetContainerIDByName(testpod, "test")
+			Expect(err).ToNot(HaveOccurred())
+
+			containersCgroups = strings.Trim(containersCgroups, "\n")
+			containersCgroupsDirs := strings.Split(containersCgroups, "\n")
+			Expect(len(containersCgroupsDirs)).To(Equal(2), "unexpected amount of containers cgroups")
+
+			for _, dir := range containersCgroupsDirs {
+				// skip application container cgroup
+				if strings.Contains(dir, containerID) {
+					continue
+				}
+
+				By("Checking what CPU the infra container is using")
+				cmd = []string{"/bin/bash", "-c", fmt.Sprintf("cat %s/cpuset.cpus", dir)}
+				output, err := nodes.ExecCommandOnNode(cmd, workerRTNode)
+				Expect(err).ToNot(HaveOccurred())
+
+				cpus, err := cpuset.Parse(output)
+				Expect(err).ToNot(HaveOccurred())
+
+				Expect(cpus.ToSlice()).To(Equal(reservedCPUSet.ToSlice()))
 			}
 		})
 	})
@@ -382,3 +500,44 @@ func getStressPod(nodeName string) *corev1.Pod {
 		},
 	}
 }
+
+func getTestPodWithAnnotations(annotations map[string]string) *corev1.Pod {
+	testpod := pods.GetTestPod()
+	testpod.Annotations = annotations
+	testpod.Namespace = testutils.NamespaceTesting
+
+	cpus := resource.MustParse("1")
+	memory := resource.MustParse("256Mi")
+
+	// change pod resource requirements, to change the pod QoS class to guaranteed
+	testpod.Spec.Containers[0].Resources = corev1.ResourceRequirements{
+		Limits: corev1.ResourceList{
+			corev1.ResourceCPU:    cpus,
+			corev1.ResourceMemory: memory,
+		},
+	}
+
+	runtimeClassName := components.GetComponentName(profile.Name, components.ComponentNamePrefix)
+	testpod.Spec.RuntimeClassName = &runtimeClassName
+	testpod.Spec.NodeSelector = map[string]string{testutils.LabelHostname: workerRTNode.Name}
+
+	return testpod
+}
+
+func deleteTestPod(testpod *corev1.Pod) {
+	// it possible that the pod already was deleted as part of the test, in this case we want to skip teardown
+	key := types.NamespacedName{
+		Name:      testpod.Name,
+		Namespace: testpod.Namespace,
+	}
+	err := testclient.Client.Get(context.TODO(), key, testpod)
+	if errors.IsNotFound(err) {
+		return
+	}
+
+	err = testclient.Client.Delete(context.TODO(), testpod)
+	Expect(err).ToNot(HaveOccurred())
+
+	err = pods.WaitForDeletion(testpod, pods.DefaultDeletionTimeout*time.Second)
+	Expect(err).ToNot(HaveOccurred())
+}
